# distutils: language = c++

# Copyright 2017 neurodata (http://neurodata.io/)
# Written by Disa Mhembere (disa@jhu.edu)
#
# This file is part of knor.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp.vector cimport vector
from libcpp.string cimport string
import numpy as np
cimport numpy as np
import ctypes
import sys
from os.path import abspath
from exceptions import NotImplementedError, RuntimeError
from Exceptions.runtime import UnsupportedError

# Metadata
__version__ = "0.0.1"
__author__ = "Disa Mhembere"
__maintainer__ = "Disa Mhembere <disa@jhu.edu>"
__package__ = "knor"

cdef extern from "kmeans_types.hpp" namespace "kpmeans::base":
    cdef cppclass kmeans_t:
        kmeans_t()
        kmeans_t(const size_t, const size_t, const size_t,
                 const size_t, const unsigned*,
                 const size_t*, const vector[double]&)
        const void write(const string dirname) const
        void set_params(const size_t nrow, const size_t ncol,
                const size_t iters, const size_t k)
        size_t nrow
        size_t ncol
        size_t iters
        size_t k
        vector[unsigned] assignments
        vector[size_t] assignment_count
        vector[double] centroids

cdef class knor_t:
    cdef kmeans_t c_kmeans_t      # hold a C++ instance which we're wrapping
    def __cinit__(self, const size_t nrow, const size_t ncol, const size_t iters,
             const size_t k, vector[unsigned]& assignments_buf,
             const vector[size_t]& assignment_count_buf,
             const vector[double]& centroids):
        self.c_kmeans_t = kmeans_t(nrow, ncol, iters, k, assignments_buf.data(),
                assignment_count_buf.data(), centroids)

    def __repr__(self):
        s = "Iterations: {}\nk: {}\nSizes:\n{}\nCentroids:\n{}\nCluster:\n{}\n"
        return s.format(self.get_iters(), self.get_k(),
                        self.get_sizes(), self.get_centroids(),
                        self.get_clusters())

    def write(self, const string dirname):
        return self.c_kmeans_t.write(dirname)

    def get_nrow(self):
        return self.c_kmeans_t.nrow

    def get_k(self):
        return self.c_kmeans_t.k

    def get_iters(self):
        return self.c_kmeans_t.iters

    def get_ncol(self):
        return self.c_kmeans_t.ncol

    def get_clusters(self):
        return np.array(self.c_kmeans_t.assignments)

    def get_sizes(self):
        return np.array(self.c_kmeans_t.assignment_count)

    def get_centroids(self):
        centroids = np.array(self.c_kmeans_t.centroids)
        return centroids.reshape(self.get_k(), self.get_ncol())

    def __richcmp__(self, other, int op):
        if op == 2: # 2 is __eq__
            if isinstance(other, self.__class__):
                return np.array_equal(self.get_clusters(),
                        other.get_clusters()) and\
                    np.array_equal(self.get_sizes(), other.get_sizes()) and\
                    np.array_equal(self.get_centroids(),
                            other.get_centroids()) and\
                    np.array_equal(self.get_iters(), other.get_iters())
            else:
                return False

    def set_params(self, const size_t nrow, const size_t ncol,
            const size_t iters, const size_t k):
        return self.c_kmeans_t.set_params(nrow, ncol, iters, k)

cdef extern from "util.hpp" namespace "kpmeans::base":
    cpdef int get_num_omp_threads()
    cpdef unsigned get_num_nodes()

cdef extern from "knori.hpp" namespace "kpmeans::base":
    kmeans_t kmeans(double* data, const size_t nrow,
        const size_t ncol, const unsigned k, size_t max_iters,
        unsigned nnodes, unsigned nthread,
        double* p_centers, string init,
        double tolerance, string dist_type,
        bint omp, bint numa_opt)

    kmeans_t kmeans(const string datafn, const size_t nrow,
        const size_t ncol, const unsigned k,
        size_t max_iters, unsigned nnodes,
        unsigned nthread, double* p_centers, string init,
        double tolerance, string dist_type, bint omp)

# Help
def build_defaults(kwargs):
    DEFAULT_ARGS = {
        "max_iters": sys.maxint, "nnodes": get_num_nodes(),
        "nthread": get_num_omp_threads(), "p_centers": None,
        "init": "kmeanspp", "tolerance": -1, "dist_type": "eucl",
        "omp": False, "numa_opt": False}

    for k in DEFAULT_ARGS.iterkeys():
        if k not in kwargs:
            kwargs[k] = DEFAULT_ARGS[k]

def read(fn, vector[double]& v):
    f = open(fn, "rb")
    CHUNKSIZE = 8
    try:
        while True:
            data_read = file.read(CHUNKSIZE)
            if data_read:
                v.push_back(<double>data_read)
            else:
                break
    finally:
        f.close()

def Kmeans(np.ndarray[double, ndim=2] data, centers, **kwargs):
    """
    Run the k-means algorithm for data on the local file system.
    K-means provides *k* disjoint sets for a dataset using a parallel
    and fast NUMA optimized version of Lloyd's algorithm. The details
    of which are found in this paper https://arxiv.org/pdf/1606.08905.pdf.

    **Positional Arguments:**

        data:
            - A path to a file containing a binary row-major double floating
            point precision "matrix".
        centers:
            - Either (i) The number of centers (i.e., k), or (ii) an
            In-memory data matrix (numpy.ndarray)

    **Keyword Arguments:**

        max_iters:
            - The maximum number of iteration of k-means to perform
            (Default: System Max)
        nthread:
            - The number of parallel thread to run (Default: # Phys Cores)
        init:
            - The type of initialization to use "kmeanspp", "random",
                "forgy", "none" (Default: "kmeanspp")
        tolerance:
            - The convergence tolerance (Default: 0)
        dist_type: What dissimilarity metric to use "eucl", "cos" (Default:
        "eucl")
        omp:
            - Use (slower) OpenMP threads rather than pthreads` for
            parallelizations.
        numa_opt:
            - When passing `data` as an in-memory data matrix you can
            optimize memory placement for Linux NUMA machines. *NOTE:*
            performance may degrade with very large data & it
            requires 2*memory of that without this.
    """

    nrow = data.shape[0]
    ncol = data.shape[1]
    build_defaults(kwargs)
    max_iters = kwargs["max_iters"]
    nnodes = kwargs["nnodes"]
    nthread = kwargs["nthread"]
    cdef p_centers = kwargs["p_centers"]
    init = kwargs["init"]
    tolerance = kwargs["tolerance"]
    dist_type = kwargs["dist_type"]
    omp = kwargs["omp"]
    numa_opt = kwargs["numa_opt"]

    # C order only
    if np.isfortran(data):
        data = data.transpose()

    cdef kmeans_t ret
    cdef vector[double] c_centers

    # Centers computed
    if isinstance(centers, int) or isinstance(centers, long):
        ret = kmeans(&data[0,0], nrow, ncol,
                centers, max_iters, nnodes, nthread, NULL,
                init, tolerance, dist_type, omp, numa_opt)

    # Centers in-memory
    elif isinstance(centers, np.ndarray):
        if np.isfortran(centers):
            centers = centers.transpose()

        for item in centers.flatten():
            c_centers.push_back(item)

        ret = kmeans(&data[0,0], nrow, ncol, centers.shape[0], max_iters,
                nnodes, nthread, &c_centers[0],
                "none", tolerance, dist_type, omp, numa_opt)
    else:
        raise UnsupportedError("centers must be of type `long/int` or " +\
            "`numpy.ndarray`\n")

    return knor_t(ret.nrow, ret.ncol, ret.iters, ret.k, ret.assignments,
            ret.assignment_count, ret.centroids)

def KmeansEM(string data, centers, nrow, ncol, **kwargs):
    """
    Run the k-means algorithm for data on the local file system.
    K-means provides *k* disjoint sets for a dataset using a parallel
    and fast NUMA optimized version of Lloyd's algorithm. The details
    of which are found in this paper https://arxiv.org/pdf/1606.08905.pdf.

    **Positional Arguments:**

        data:
            - A path to a file containing a binary row-major double floating
            point precision "matrix".
        centers:
            - Either (i) The number of centers (i.e., k), or (ii) an
            In-memory data matrix (numpy.ndarray)
        nrow:
            - The number of rows in the data
        ncol:
            - The number of columns in the data

    **Keyword Arguments:**

        max_iters:
            - The maximum number of iteration of k-means to perform
            (Default: System Max)
        nthread:
            - The number of parallel thread to run (Default: # Phys Cores)
        init:
            - The type of initialization to use "kmeanspp", "random",
                "forgy", "none" (Default: "kmeanspp")
        tolerance:
            - The convergence tolerance (Default: 0)
        dist_type: What dissimilarity metric to use "eucl", "cos" (Default:
        "eucl")
        omp:
            - Use (slower) OpenMP threads rather than pthreads` for
            parallelizations.
        numa_opt:
            - When passing `data` as an in-memory data matrix you can
            optimize memory placement for Linux NUMA machines. *NOTE:*
            performance may degrade with very large data & it
            requires 2*memory of that without this.
    """

    build_defaults(kwargs)
    max_iters = kwargs["max_iters"]
    nnodes = kwargs["nnodes"]
    nthread = kwargs["nthread"]
    cdef p_centers = kwargs["p_centers"]
    init = kwargs["init"]
    tolerance = kwargs["tolerance"]
    dist_type = kwargs["dist_type"]
    omp = kwargs["omp"]

    if not nrow or not ncol:
        raise RuntimeError("Data dim cannot be 0: `nrow` or `ncol`")

    cdef kmeans_t ret
    cdef vector[double] c_centers

    # Centers computed
    if isinstance(centers, int) or isinstance(centers, long):
        ret = kmeans(abspath(data), nrow, ncol,
                centers, max_iters, nnodes, nthread, NULL,
                init, tolerance, dist_type, omp)

    # Centers in-memory
    elif isinstance(centers, np.ndarray):
        if np.isfortran(centers):
            centers = centers.transpose()

        for item in centers.flatten():
            c_centers.push_back(item)

        ret = kmeans(abspath(data), nrow, ncol, centers.shape[0], max_iters,
                nnodes, nthread, &c_centers[0],
                "none", tolerance, dist_type, omp)
    else:
        raise UnsupportedError("centers must be of type `long/int` or " +\
            "`numpy.ndarray`\n")

    return knor_t(ret.nrow, ret.ncol, ret.iters, ret.k, ret.assignments,
            ret.assignment_count, ret.centroids)



#!/usr/bin/env python

import os
import sys, re
from glob import glob
from exceptions import NotImplementedError
from exceptions import RuntimeError
from distutils.errors import DistutilsSetupError
from distutils.command.build_clib import build_clib
from distutils.core import setup, Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext

_REPO_ISSUES_ = "https://github.com/flashxio/knorPy/issues"
_OS_SUPPORTED_ = {"linux":"linux", "mac":"darwin"}


patts = []
for opsys in _OS_SUPPORTED_.itervalues():
    patts.append(re.compile("(.*)("+opsys+")(.*)"))

raw_os = sys.platform.lower()

OS = None
for patt in patts:
    res = re.match(patt, raw_os)
    if res is not None:
        OS = res.groups()[1]
        break

################################################################################

class knor_clib(build_clib, object):
    def initialize_options(self):
        super(knor_clib, self).initialize_options()
        self.include_dirs = [
            "libman", "binding", "libkcommon",
            "/usr/local/opt/boost/include",
            "/usr/local/lib/python2.7/site-packages/numpy/core/include",
            ]

        self.define = [
                ("BOOST_LOG_DYN_LINK", None),
                ("BIND", None), ("OSX", None)
                ]

    def build_libraries(self, libraries):
        for (lib_name, build_info) in libraries:
            sources = build_info.get('sources')
            if sources is None or not isinstance(sources, (list, tuple)):
                raise DistutilsSetupError, \
                      ("in 'libraries' option (library '%s'), " +
                       "'sources' must be present and must be " +
                       "a list of source filenames") % lib_name
            sources = list(sources)

            print "building '%s' library" % lib_name

            # First, compile the source code to object files in the library
            # directory.  (This should probably change to putting object
            # files in a temporary build directory.)
            macros = build_info.get('macros')
            include_dirs = build_info.get('include_dirs')

            # pass flasgs to compiler
            extra_preargs = ["-std=c++11", "-O3", "-Wno-unused-function"]

            objects = self.compiler.compile(sources,
                                            output_dir=self.build_temp,
                                            macros=macros,
                                            include_dirs=include_dirs,
                                            debug=self.debug,
                                            extra_preargs=extra_preargs)

            # Now "link" the object files together into a static library.
            # (On Unix at least, this isn't really linking -- it just
            # builds an archive.  Whatever.)
            self.compiler.create_static_lib(objects, lib_name,
                                            output_dir=self.build_clib,
                                            debug=self.debug)

################################################################################

# For C++ libraries
libkcommon = ("kcommon",
        {"sources": glob(os.path.join("libkcommon/", "*.cpp"))})
libman = ('man',
        {'sources': glob(os.path.join("libman/", "*.cpp"))})

sources = glob(os.path.join("binding/", "*.cpp"))
sources.append("python/knor/knor.pyx")

ext_modules = cythonize(Extension(
        "knor.knor",                                # the extension name
        sources=sources,
        language="c++",
        extra_compile_args=["-std=c++11", "-O3",
            "-Wno-unused-function", "-I.","-Ilibman",
            "-Ibinding", "-Ilibkcommon",
            "-I/usr/local/lib/python2.7/site-packages/numpy/core/include",
            "-DBOOST_LOG_DYN_LINK", "-I/usr/local/opt/boost/include",
            "-DBIND", "-DOSX"],
        extra_link_args=[
            "-Llibman", "-lman", "-Llibkcommon",
            "-lkcommon", "-lpthread", "-lboost_log-mt",
            "-lboost_system", "-L/usr/local/opt/boost/lib",
            ]))

if OS is None:
    raise RuntimeError("Operating system {}\n." +\
            "Please post an issue at {}\n".format(raw_os, _REPO_ISSUES_))

elif OS == _OS_SUPPORTED_["linux"]:
    raise NotImplementedError("Linux OS support")
elif OS == _OS_SUPPORTED_["mac"]:

    setup(
        name="knor",
        version="0.0.1a8",
        description="A fast parallel k-means library for Linux and Mac",
        long_description="The k-means NUMA Optimized Routine library or " +\
        "knor is a highly optimized and fast library for computing " +\
        "k-means in parallel with accelerations for Non-Uniform Memory " +\
        "Access (NUMA) architectures",
        url="https://github.com/flashxio/knor",
        author="Disa Mhembere",
        author_email="disa@jhu.edu",
        license="Apache License, Version 2.0",
        keywords="kmeans k-means parallel clustering machine-learning",
        install_requires=[
            "numpy",
            "Cython==0.23.5",
            "cython==0.23.5",
            ],
        package_dir = {"knor": os.path.join("python", "knor")},
        packages=["knor", "knor.Exceptions"],
        libraries = [libkcommon, libman],
        cmdclass = {'build_clib': knor_clib, 'build_ext': build_ext},
        ext_modules = ext_modules,
        )
else:
    assert False, "Unsupported OS NOT correctly caught by knor"
